// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#ifndef __spatialindex_btree_btree_h
#define __spatialindex_btree_btree_h

#include "Statistics.h"
#include "Node.h"
#include "PointerPoolNode.h"



namespace SpatialIndex
{
	namespace BTree
	{
		class BTree : public ISpatialIndex
		{

		public:
			BTree(IStorageManager&, Tools::PropertySet&);
				// String                   Value     Description
				// ----------------------------------------------
				// IndexIndentifier         VT_LONG   If specified an existing index will be openened from the supplied
				//                          storage manager with the given index id. Behaviour is unspecified
				//                          if the index id or the storage manager are incorrect.
				// Dimension                VT_ULONG  Dimensionality of the data that will be inserted.
				// IndexCapacity            VT_ULONG  The index node capacity. Default is 100.
				// LeafCapactiy             VT_ULONG  The leaf node capacity. Default is 100.
				// FillFactor               VT_DOUBLE The fill factor. Default is 70%
				// TreeVariant              VT_LONG   Can be one of B+ 
				// IndexPoolCapacity        VT_LONG   Default is 100
				// LeafPoolCapacity         VT_LONG   Default is 100
				// PointPoolCapacity        VT_LONG   Default is 500
				// EmbeddingFanout	    VT_LONG   Default is 2
			
			virtual ~BTree();

			//
			// ISpatialIndex interface
			//
			virtual void insertData(unsigned long len, const byte* pData, const IShape& shape, long shapeIdentifier) throw (Tools::Exception, std::exception);
			virtual bool deleteData(const IShape& shape, long id) throw (Tools::Exception, std::exception);
			virtual void containsWhatQuery(const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception);
			virtual void intersectsWithQuery(const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception);
			virtual void pointLocationQuery(const Point& query, IVisitor& v) throw (Tools::Exception, std::exception);
			virtual void nearestNeighborQuery(long k, const IShape& query, IVisitor& v, INearestNeighborComparator&) throw (Tools::Exception, std::exception);
			virtual void nearestNeighborQuery(long k, const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception);
			virtual void selfJoinQuery(const IShape& s, IVisitor& v) throw (Tools::Exception, std::exception);
			virtual void queryStrategy(IQueryStrategy& qs) throw (Tools::Exception, std::exception);
			virtual void getIndexProperties(Tools::PropertySet& out) const;
			virtual void addCommand(ICommand* pCommand, CommandType ct);
			virtual bool isIndexValid() throw (Tools::Exception, std::exception);
			virtual void getStatistics(IStatistics** out) const throw (std::exception);
			virtual void showBTree(NodePtr root);
			virtual void showTree();
			virtual void computeFanout();
			virtual void computeHashes();
			virtual void startCountingAffectedNodes();
			virtual void stopCountingAffectedNodes();
			virtual void clearCountAffectedNodes();

			
		private:
			void initNew(Tools::PropertySet&);
			void initOld(Tools::PropertySet& ps);
			void storeHeader();
			void loadHeader();

			bool deleteData_impl(const Point& key);

			long writeNode(Node*);
			NodePtr readNode(unsigned long);
			void deleteNode(Node*);

			void rangeQuery(RangeQueryType type, const IShape& query, IVisitor& v) throw (Tools::Exception, std::exception);
			
			long compareKeys(long key1, long key2);
			//search
			NodePtr descendToLeaf(long searchKey, NodePtr curr, IVisitor& v);
			NodePtr search(long searchKey, IVisitor& v);
			//insert
			void insert (Point& key, long id);
			NodePtr descendSplit(Point& key, long id, NodePtr curr, NodePtr parent);
			void makeNewRoot(Point& key, NodePtr oldRoot, NodePtr newNode);
			NodePtr split(NodePtr curr);
			//deletion
			NodePtr descendBalance(const Point& key, NodePtr curr, NodePtr left, NodePtr right, NodePtr lAnc, NodePtr rAnc, NodePtr parent);
			NodePtr shift(NodePtr left, NodePtr right, NodePtr anchor);
			NodePtr merge(NodePtr left, NodePtr right, NodePtr anchor);
			void collapseRoot(NodePtr oldRoot, NodePtr newRoot);
			void updateTreeDigestSig();
			void computeNodeHash(NodePtr node, byte** hash);

			
			IStorageManager* m_pStorageManager;

			long m_rootID;
			long m_headerID;
			long m_splitNodeID;
			long m_mergeNodeID;
			long m_digestUpdateNodeID;

			BTreeVariant m_treeVariant;

			double m_fillFactor;

			unsigned long m_indexCapacity;

			unsigned long m_leafCapacity;

			unsigned long m_dimension;
			
			unsigned long m_indexMinFanout;
			unsigned long m_leafMinFanout;
		
			long m_embeddingFanout;
			
			byte* m_treeDigest;
			byte* m_treeSignature;
			unsigned long m_treeDigestLength;
			unsigned long m_treeSignatureLength;

			Point m_infinitePoint;
			Point m_tmpKey;			

			Statistics m_stats;
			clock_t m_c0,m_c1;
			double m_cputime;
			long m_hashComputed;
			unsigned long m_avgFanout;
			unsigned long m_avgIndexFanout;
			unsigned long m_rootFanout;
			unsigned long m_totalNode;
			std::map<long, long> m_affectedNodes;
			bool m_countAffectedNodes;


			Tools::PointerPool<Point> m_pointPool;
			Tools::PointerPool<Node> m_indexPool;
			Tools::PointerPool<Node> m_leafPool;

			std::vector<Tools::SmartPointer<ICommand> > m_writeNodeCommands;
			std::vector<Tools::SmartPointer<ICommand> > m_readNodeCommands;
			std::vector<Tools::SmartPointer<ICommand> > m_deleteNodeCommands;

#ifdef PTHREADS
			pthread_rwlock_t m_rwLock;
#else
			bool m_rwLock;
#endif


			friend class Node;
			friend class Leaf;
			friend class Index;

			friend std::ostream& operator<<(std::ostream& os, const BTree& t);
		}; // BTree

		std::ostream& operator<<(std::ostream& os, const BTree& t);
	}
}

#endif /*__spatialindex_btree_btree_h*/

