// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#ifndef __spatialindex_btree_node_h
#define __spatialindex_btree_node_h


namespace SpatialIndex
{
	namespace BTree
	{
		class BTree;
		class Leaf;
		class Index;
		class Node;

		typedef Tools::PoolPointer<Node> NodePtr;

		class Node : public SpatialIndex::INode
		{
		public:
			virtual ~Node();

			//
			// Tools::IObject interface
			//
			virtual Tools::IObject* clone() throw (Tools::NotSupportedException);

			//
			// Tools::ISerializable interface
			//
			virtual unsigned long getByteArraySize();
			virtual unsigned long loadFromByteArray(byte* const data);
			virtual void storeToByteArray(unsigned long& len, byte** data);

			//
			// SpatialIndex::IEntry interface
			//
			virtual long getIdentifier() const;
			virtual void getShape(IShape** out) const throw (std::exception);

			//
			// SpatialIndex::INode interface
			//
			virtual unsigned long getChildrenCount() const;
			virtual long getChildIdentifier(unsigned long index)  const throw (Tools::IndexOutOfBoundsException);
			virtual void getChildShape(unsigned long index, IShape** out)  const throw (Tools::IndexOutOfBoundsException, std::exception);
			virtual unsigned long getLevel() const;
			virtual bool isIndex() const;
			virtual bool isLeaf() const;

		private:
			Node();
			Node(BTree* pTree, long id, unsigned long level, unsigned long capacity);

			virtual Node& operator=(const Node&);

			virtual void deleteEntry(unsigned long index);

			virtual NodePtr chooseSubtree(const Point& key, unsigned long level, std::stack<long>& pathBuffer) = 0;
			virtual NodePtr findLeaf(const Point& key, long id, std::stack<long>& pathBuffer) = 0;
			virtual long getSlot(long searchKey);	
			virtual long findKey(long searchkey, long lo, long hi);
			virtual long bestMatch(long searchKey, long slot);
			virtual bool isFull();
			virtual bool isFewest();
			virtual void insertEntry(long slot, Point& key, long id, NodePtr sibling, NodePtr downPtr);
			virtual void placeEntry(long slot, Point& key, long id, NodePtr downPtr);
			virtual void showNode() = 0;
			virtual void removeEntry(long slot);
			virtual void computeNodeDigests(NodePtr node, byte* digest);
			virtual void getkeydigArray(long** keys,long** pids, int** ids,  byte**& digests);
			virtual void buildEmbeddings();
			virtual void inorderTraverse(RBNode* node, long& pos);
			void restoreEmbeddingTree();
			RBNode* restoreRBNode(int pos);
			void updateEmbeddingTree();


			
			BTree* m_pTree;
				// Parent of all nodes.

			unsigned long m_level;
				// The level of the node in the tree.
				// Leaves are always at level 0.

			long m_identifier;
				// The unique ID of this node.

			unsigned long m_children;
				// The number of children pointed by this node.

			unsigned long m_capacity;
				// Specifies the node capacity.

			unsigned long m_minFanout;
				//Specifies the node's minimum fanout.

			byte** m_pDigest;
				//Hashes or Signatures associated with each key.

			unsigned long m_embeddingLength;
	
			PointPtr* m_ptrKey;	
				// The corresponding keys

			long* m_pIdentifier;
				// The corresponding data identifiers, i.e. pointers to children nodes.
	
			int** m_pID;
				// G-way search tree's pointer

			byte*** m_pHash;
				//G-way search tree's hashes
			
			int m_rootID;
				//G-way tree's root
			
			byte* m_treeDigest;
				//G-way tree's digest

			long m_pFirstNode;
				// The pointer to the first child node for index node, or the pointer to the next node in leaf node.
			
			byte* m_pFirstDigest;
				// For index node, this one stores the digest for the child pointed by first pointer
		
			RBTree* m_embeddingTree;
				// The embedding tree;

			// Needed to access protected members without having to cast from Node.
			// It is more efficient than using member functions to access protected members.
			friend class BTree;
			friend class Leaf;
			friend class Index;
			friend class Tools::PointerPool<Node>;
			friend class BulkLoader;
		}; // Node
	}
}

#endif /*__spatialindex_btree_node_h*/
